<?php

namespace Drupal\rift;

use Drupal\combined_image_style\Entity\CombinedImageStyle;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\file\Entity\File;
use Drupal\image\Entity\ImageStyle;
use Drupal\image\ImageEffectManager;
use Drupal\image\ImageStyleInterface;
use Drupal\media\MediaInterface;
use Drupal\rift\Html\ImgElement;
use Drupal\rift\Html\PictureElement;
use Drupal\rift\Html\SizesItem;
use Drupal\rift\Html\SourceElement;
use Drupal\rift\Html\SrcSetItem;

/**
 * Service to generate responsive picture markup from a media entity.
 */
class RiftPicture {

  /**
   * The module handler to invoke alter hooks.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The theme manager to invoke alter hooks.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected $themeManager;

  /**
   * Drupal Renderer Service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * File Url Generator Service.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected $fileUrlGenerator;

  /**
   * Image Effect Plugin Manager.
   *
   * @var \Drupal\image\ImageEffectManager
   */
  protected $pluginManagerImageEffect;

  /**
   * Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Stores overall configuration.
   *
   * @var \Drupal\rift\PictureConfig
   */
  protected PictureConfig $pictureConfig;

  protected RiftMedia $riftMedia;

  /**
   * Constructs the TwigTweakExtension object.
   */
  public function __construct(
    ModuleHandlerInterface $module_handler,
    ThemeManagerInterface $theme_manager,
    RendererInterface $renderer,
    FileUrlGeneratorInterface $file_url_generator,
    ImageEffectManager $plugin_manager_image_effect,
    EntityTypeManager $entity_type_manager,
    ConfigFactoryInterface $config_factory,
    RiftMedia $rift_media
  ) {
    $this->moduleHandler = $module_handler;
    $this->themeManager = $theme_manager;
    $this->renderer = $renderer;
    $this->fileUrlGenerator = $file_url_generator;
    $this->pluginManagerImageEffect = $plugin_manager_image_effect;
    $this->entityTypeManager = $entity_type_manager;
    $this->configFactory = $config_factory;
    $this->pictureConfig = new PictureConfig();
    $this->riftMedia = $rift_media;
  }

  /**
   * Generate picture sources from a given configuration and media entity.
   *
   * @param \Drupal\rift\PictureConfig $config
   *   Picture configuration.
   * @param \Drupal\media\MediaInterface $media
   *   The media entity.
   *
   * @return array
   *   Render array for responsive picture.
   */
  protected function generatePictureSources(PictureConfig $config, MediaInterface $media) {
    $picture_sources_data = [];
    $image_transforms = [];
    $ndx = 0;
    foreach ($config->sizes as $size) {
      $aspect_ratio_style = $config->aspectRatios[$ndx] ?: '100x1';
      [
        $screen,
        $view_width,
        $width,
      ] = $this->extractWidthFromScreenAndViewWidth($size, $config->screens);

      $transform = [
        'size' => $size,
        'screen' => $screen,
        'screen_width' => $config->screens[$screen]->width,
        'view_width' => $view_width,
        'width' => $width,
        'media' => $config->screens[$screen]->mediaQuery,
        'media_picture_source' => !isset($config->sizes[$ndx + 1]) ? FALSE : $config->screens[$screen]->mediaQuery,
        'aspect_ratio_style' => $aspect_ratio_style,
        'transform_style' => !empty($config->transforms[$ndx]) ? $config->transforms[$ndx] : FALSE,
      ];
      $ndx++;
      $image_transforms[] = $transform;
    }

    foreach ($config->formats as $format) {
      foreach ($image_transforms as $t) {
        $view_width = $t['view_width'];
        $width = intval($t['width']);
        $srcsets = [];
        foreach ($config->multipliers as $multiplier) {
          $quality = $config->quality[$multiplier];
          $m = floatval(trim($multiplier, 'x'));
          $transformed_width = intval($width * $m);
          $srcset = [
            'styles' => [
              'image_boundaries' => $this->generateImageBoundaryStyle($t['aspect_ratio_style'], $transformed_width, $media),
              'main' => $t['transform_style'] ?: FALSE,
              'format' => $format == 'original' ? FALSE : $format,
              'quality' => $quality ?: FALSE,
            ],
            'multiplier' => $multiplier,
            'type' => $format == 'original' ? FALSE : 'image/' . $format,
          ];
          // @todo If media is focalpoint based, use fp scale and crop.
          $srcsets[] = $srcset;
        }
        $picture_sources_data[] = [
          'media' => $t['media_picture_source'],
          'sizes' => $view_width,
          'view_width' => $view_width,
          'srcset' => $srcsets,
          'width' => intval($t['width']),
        ];
      }
    }
    return $picture_sources_data;
  }

  /**
   * Generating boundary defining image sylte for a given parameters.
   *
   * @param string $aspect_ratio_style
   *   The aspect ratio for the image to be used.
   * @param int $width
   *   The constraining width configuration.
   * @param \Drupal\media\MediaInterface $media
   *   The media entity.
   *
   * @return string
   *   Image style/effect to be used.
   */
  protected function generateImageBoundaryStyle(string $aspect_ratio_style, int $width, MediaInterface $media): string {

    if ($aspect_ratio_style && $this->riftMedia->detectIfCropExists($media, $aspect_ratio_style)) {
      return $aspect_ratio_style . '-' . $width . 'w';
    }
    else {
      assert(!empty($aspect_ratio_style), 'Aspect ratio must not be empty.');
      [$arw, $arh] = explode('x', $aspect_ratio_style);
      $arh_int = intval($arh);
      $arw_int = intval($arw);
      assert(!empty($arh_int), 'Invalid height found for aspect ratio.');
      assert(!empty($arw_int), 'Invalid width found for aspect ratio.');
      try {
        $aspect_ratio = intval($arw) / intval($arh);
        $height = intval($width / $aspect_ratio);
      }
      catch (\DivisionByZeroError $exception) {
        // Use a 4:3 aspect ratio as a failsafe.
        $height = intval($width/(4 / 3));
      }
      return $width . 'w' . $height . 'h';
    }
  }

  /**
   * Responsive Picture twig filter.
   *
   * @param array|null $element
   *   Render array element.
   * @param array $config
   *   Configuration for generating srcset.
   *   <code>
   *   content.field_common_image|repsonsive_image({
   *   sources: "120w 320w 320w 1024w 1024w",
   *   ratios: "1x1 16x9 1x1 1x1 1x1",
   *   quality: {
   *   1x: "80 80 80 80 80",
   *   2x: "40 40 40 40 40",
   *   },
   *   types: {webp avif jpeg},
   *   })
   *   </code>.
   *
   * @return string[]
   *   Render array for responsive picture tag.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function responsivePictureFromElement($element, array $config = []): array {
    // @todo Instead of bubbleUpCacheTags, inject proper cache context.
    // Cache context varies based on RULE and Media entity.
    $this->bubbleUpCacheTags($element);
    $media = $this->getImageMedia($element);
    if ($media instanceof MediaInterface) {
      return $this->responsivePicture($media, $config);
    }
    return [
      '#markup' => '<!-- Missing Image -->',
    ];
  }

  /**
   * Get sitewide default configurations.
   *
   * @return array
   *   The default configurations.
   */
  protected function getDefaultConfig(): array {
    return $this->configFactory->get('rift.settings')->get('config');
  }

  /**
   * Responsive Picture twig filter.
   *
   * @param \Drupal\media\MediaInterface $media
   *   Media entity.
   * @param array $config
   *   Configuration for generating srcset.
   *   <code>
   *   content.field_common_image|repsonsive_image({
   *   sources: "120w 320w 320w 1024w 1024w",
   *   ratios: "1x1 16x9 1x1 1x1 1x1",
   *   quality: {
   *   1x: "80 80 80 80 80",
   *   2x: "40 40 40 40 40",
   *   },
   *   types: {webp avif jpeg},
   *   })
   *   </code>.
   *
   * @return string[]
   *   Render array for responsive picture tag.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function responsivePicture(MediaInterface $media, array $config = []): array {
    $image_field = $uri = NULL;
    if ($file = $this->getImageFileEntity($media)) {
      $image_field = $this->getImageField($media)[0];
      if ($file instanceof File) {
        $uri = $file->getFileUri();
      }
    }
    if (empty($image_field) || empty($uri)) {
      // @todo Check if we can provide placeholder images.
      return [
        '#markup' => '<!-- Missing Image -->',
      ];
    }

    $config = array_merge($this->getDefaultConfig(), $config);

    $this->pictureConfig
      ->processScreens($config['screens'])
      ->processSizes($config['sizes'])
      ->processTransforms($config['transforms'])
      ->processMultipliers($config['multipliers'])
      ->processQuality($config['quality'])
      ->processAspectRatios($config['aspect_ratios'])
      ->processFormats($config['formats'])
      ->processAttribute($config['attributes'])
      ->processFallbackTransforms($config['fallback_transform'])
      ->processUrlGenerationStrategy($config['url_generation_strategy']);

    if (empty($this->pictureConfig->formats)) {
      return [
        '#markup' => '<!-- Missing Formats -->',
      ];
    }
    if (empty($this->pictureConfig->screens)) {
      return [
        '#markup' => '<!-- Missing screens -->',
      ];
    }
    $sources = $this->generatePictureSources($this->pictureConfig, $media);

    if (empty($sources)) {
      // @todo Check if we can provide placeholder images.
      return [
        '#markup' => '<!-- Missing Image -->',
      ];
    }

    $picture_element = new PictureElement();
    $source_elements = [];
    foreach ($sources as $source) {
      $source_elements[] = $this->generateSourceTagFromStyleArray($source, $uri);
    }
    $picture_element->setSource($source_elements);

    // Generate the <img> tag inside <picture>.
    $image_data = $image_field->getValue();
    $combinedImageStyle = (new CombinedImageStyle())
      ->setSourceUri($uri)
      ->setImageStyles(explode('-', $this->pictureConfig->fallbackTransform));
    $url = $combinedImageStyle->buildCombinedUrl();

    $img_element = new ImgElement();
    $img_element
      ->setSrc($url)
      ->setAlt($image_data['alt'])
      ->setTitle($image_data['title']);
    $picture_element->setImg($img_element);
    $picture_element->setAttribute($this->pictureConfig->attribute);
    return $picture_element->render();
  }

  /**
   * Detect Image file uri from given render element.
   *
   * @param array|null $element
   *   Render array element.
   *
   * @return \Drupal\media\MediaInterface|null
   *   Uri if detected, otherwise null.
   */
  protected function getImageMedia($element): ?MediaInterface {
    $children = array_intersect_key($element, array_flip(Element::children($element)));
    $media = $children[0]['#media'] ?: NULL;
    if ($media instanceof MediaInterface) {
      $source = $media->getSource();
      if ($source->getPluginId() === 'image') {
        return $media;
      }
    }
    return NULL;
  }

  /**
   * Get Image field from given render array element.
   *
   * @param \Drupal\media\MediaInterface $media
   *   A render array element.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   Media entity if present, otherwise null.
   */
  protected function getImageFileEntity(MediaInterface $media): ?EntityInterface {
    $field_items = $this->getImageField($media);
    if ($field_items instanceof EntityReferenceFieldItemListInterface) {
      return $field_items->referencedEntities()[0] ?? NULL;
    }
    return NULL;
  }

  /**
   * Get Image field from given render array element.
   *
   * @param \Drupal\media\MediaInterface $media
   *   A render array element.
   *
   * @return \Drupal\Core\Field\FieldItemListInterface
   *   Media entity if present, otherwise null.
   */
  protected function getImageField(MediaInterface $media): FieldItemListInterface {
    $field_name = $this->riftMedia->getMediaFieldName($media);
    return $media->get($field_name);
  }

  /**
   * Get Image style from given source.
   */
  protected function getImageStyleFromSource($source): ?ImageStyleInterface {
    $style_id = $this->getImageStyleIdFromSource($source);
    $image_style = $this->entityTypeManager
      ->getStorage('image_style')
      ->load($style_id);
    if (!($image_style instanceof ImageStyleInterface)) {
      $image_style = $this->generateImageStyle($source, FALSE);
    }
    return $image_style ?: NULL;
  }

  /**
   * Generates an image style based on source definition.
   *
   * @param string $source
   *   The source definition.
   * @param bool $check_width
   *   Validate against width. Throws error when width can't be determined.
   *
   * @return \Drupal\image\ImageStyleInterface|null
   *   Image style.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function generateImageStyle(string $source, bool $check_width = TRUE): ?ImageStyleInterface {
    $id = $this->getImageStyleIdFromSource($source);
    $height = $this->getHeightFromSource($source);
    $width = $this->getWidthFromSource($source);
    if ($check_width && empty($width)) {
      trigger_error('Width cannot be empty', E_USER_WARNING);
      return NULL;
    }
    /** @var \Drupal\image\ImageStyleInterface $style */
    $style = ImageStyle::create(['name' => $id, 'label' => $id]);
    $type = 'image_scale';
    if ($height) {
      if ($this->moduleHandler->moduleExists('focal_point')) {
        $type = 'focal_point_scale_and_crop';
      }
      else {
        $type = 'image_scale_and_crop';
      }
    }
    $configuration = [
      'id' => $type,
      'uuid' => NULL,
      'weight' => 1,
      'data' => [
        'upscale' => FALSE,
        'width' => $width,
        'height' => $height,
      ],
    ];
    $effect = $this->pluginManagerImageEffect->createInstance($configuration['id'], $configuration);
    $style->addImageEffect($effect->getConfiguration());

    $extension = $this->getFileExtensionFromSource($source);
    if ($extension) {
      $configuration = [
        'id' => 'image_convert',
        'uuid' => NULL,
        'weight' => 2,
        'data' => [
          'extension' => $extension,
        ],
      ];
      $effect = $this->pluginManagerImageEffect->createInstance($configuration['id'], $configuration);
      $style->addImageEffect($effect->getConfiguration());
    }
    $style->save();
    return $style;
  }

  /**
   * Detect file extension from given source.
   *
   * @param string $source
   *   The source configuration.
   *
   * @return mixed|string|null
   *   The file extension if present.
   */
  protected function getFileExtensionFromSource($source) {
    return pathinfo($source)['extension'] ?? NULL;
  }

  /**
   * Detect height settings from source config.
   *
   * @param string $source
   *   The source configuration.
   *
   * @return mixed|null
   *   Height if present, NULL otherwise.
   */
  protected function getHeightFromSource($source) {
    preg_match('/(\d+)h/', $source, $matches, PREG_OFFSET_CAPTURE, 0);
    return empty($matches[1][0]) ? NULL : $matches[1][0];
  }

  /**
   * Detect width settings from source config.
   *
   * @param string $source
   *   The source configuration.
   *
   * @return mixed|null
   *   Width if present, NULL otherwise.
   */
  protected function getWidthFromSource($source) {
    preg_match('/(\d+)w/', $source, $matches, PREG_OFFSET_CAPTURE, 0);
    return empty($matches[1][0]) ? NULL : $matches[1][0];
  }

  /**
   * Generate Style ID from given source config.
   *
   * @param string $source
   *   The source configuration.
   *
   * @return string
   *   Machine name for style.
   */
  protected function getImageStyleIdFromSource($source) {
    $prefix = '';
    $name = strtolower($source);
    $name = preg_replace('/[^a-z0-9_]+/', '_', $name);
    $name = preg_replace('/_+/', '_', $name);
    return $prefix . $name;
  }

  /**
   * Provide bubbling up of cache tags on render elements when bypassing render.
   *
   * This would resolve issues with improper caching when field are not printed
   * as render array.
   */
  public function bubbleUpCacheTags(&$element) {
    // @todo Find proper way to do this.
    if (is_array($element) && array_key_exists('#cache', $element)) {
      // Early return if this element was pre-rendered (no need to re-render).
      // @todo Check if this case is handled by twig_capture.
      if (isset($element['#printed']) && $element['#printed'] == TRUE && isset($element['#markup']) && strlen($element['#markup']) > 0) {
        return;
      }
      show($element);
      $this->renderer->render($element);
    }
  }

  /**
   * Generate Image srcset from given sources defintion.
   *
   * @param string $sources
   *   The configurations to generate entire srcset.
   * @param string $uri
   *   The base image for which srcset needs to be generated.
   *
   * @return string
   *   Source string to be used with srcset attribute.
   */
  protected function generateImageSrcSet($sources, string $uri) {
    $srcset = [];
    foreach (explode(',', $sources) as $source) {
      $source = trim($source);
      $image_url = $this->generateImageUrlForSource($source, $uri);
      $srcset[] = $image_url . ' ' . $this->getWidthFromSource($source) . 'w';
    }
    return "\n" . implode(",\n", $srcset);
  }

  /**
   * Generate image url as defined by source for a given image uri.
   *
   * @param string $source
   *   The source config.
   * @param string $uri
   *   The Image uri for which url needs to be generated.
   *
   * @return string
   *   The Image url that can be used in 'src' or 'srcset' attributes.
   */
  protected function generateImageUrlForSource($source, $uri) {
    $image_style = $this->getImageStyleFromSource($source);
    return $this
      ->fileUrlGenerator
      ->transformRelative($image_style->buildUrl($uri));
  }

  /**
   * Process image styles identified from dynamic configuration.
   *
   * @param string[] $styles
   *   List of styles.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function processImageStyles(array $styles) {
    foreach ($styles as $style) {
      $height = $this->getHeightFromSource($style);
      $width = $this->getWidthFromSource($style);
      if ($height || $width) {
        $this->getImageStyleFromSource($style);
      }
    }
  }

  /**
   * Generate <source> tag from style definitions.
   *
   * @param mixed $config
   *   Configurations from which a source tag can be generated.
   * @param string|null $uri
   *   The uri for source tag.
   *
   * @return \Drupal\rift\Html\SourceElement
   *   Returns source tag render array.
   */
  protected function generateSourceTagFromStyleArray(mixed $config, ?string $uri): SourceElement {
    $srcset_items = [];
    $source_element = new SourceElement();
    foreach ($config['srcset'] as $source) {
      $url = $this->generateImageUrlFromStyles($source['styles'], $uri);
      $srcset_item = new SrcSetItem();
      $srcset_item->setImg($url);
      // Red Flag, why is multiplier being injected into width?
      $srcset_item->setWidth($source['multiplier']);
      $srcset_items[] = $srcset_item;
      // We set the type to the last one in the source.
      // Is this good idea?
      if ($source['type']) {
        $source_element->setType($source['type']);
      }
    }
    $sizes_item = new SizesItem();
    $sizes_item->setMediaQuery($config['media']);
    $sizes_item->setWidth($config['view_width']);
    $source_element
      ->setMedia($config['media'])
      ->setSizes([$sizes_item])
      ->setSrcset($srcset_items);
    return $source_element;
  }

  /**
   * Helper method to extract width from screen and view width definition.
   *
   * @param string $size
   *   Size definition (E.g. sm:100vw, lg:400px, ...).
   * @param \Drupal\rift\ScreenConfig[] $screens
   *   Screen configurations.
   *
   * @return array
   *   Returns screen, view_width definition and width.
   *
   * @todo fix return type.
   */
  protected function extractWidthFromScreenAndViewWidth(string $size, array $screens): array {
    [$screen, $view_width] = explode(':', $size);
    $screen_width = $screens[$screen]->width;
    $view_width = trim($view_width, 'px');
    if (str_ends_with($view_width, 'vw')) {
      $percent = floatval(trim($view_width, 'vw'));
      $width = intval($screen_width * $percent / 100);
    }
    else {
      $width = $view_width;
    }
    return [$screen, $view_width, $width];
  }

  /**
   * Generate Image url from image styles definitions.
   */
  protected function generateImageUrlFromStyles(array $input_styles, ?string $uri): string {
    return match ($this->pictureConfig->urlGenerationStrategy) {
      'combined_image_style' => $this->generateCombinedImageStyleImageUrlFromStyles($input_styles, $uri),
      'dummyimage' => $this->generateDummyImageUrlFromStyles($input_styles),
      default => $this->generatePlaceholditImageUrlFromStyles($input_styles),
    };
  }

  /**
   * Generate Image url compatible with "placehold.it" placeholder service.
   */
  protected function generatePlaceholditImageUrlFromStyles(array $input_styles): string {
    $style = implode('-', array_filter($input_styles));
    $height = $this->getHeightFromSource($style);
    $width = $this->getWidthFromSource($style);
    return 'https://place-hold.it/' . $width . 'x' . $height . '&text=' . $style;
  }

  /**
   * Generate Image url compatible with "dummy image" placeholder service.
   */
  protected function generateDummyImageUrlFromStyles(array $input_styles): string {
    $style = implode('-', array_filter($input_styles));
    $height = $this->getHeightFromSource($style);
    $width = $this->getWidthFromSource($style);
    return 'https://dummyimage.com/' . $width . 'x' . $height . '/eee/222&text=' . $style;
  }

  /**
   * Generate Image url compatible with "combined_image_style" module.
   */
  protected function generateCombinedImageStyleImageUrlFromStyles(array $input_styles, ?string $uri): string {
    $style = implode('-', array_filter($input_styles));
    $styles = explode('-', $style);
    $this->processImageStyles($styles);
    $combinedImageStyle = (new CombinedImageStyle())
      ->setSourceUri($uri)
      ->setImageStyles($styles);
    return $combinedImageStyle->buildCombinedUrl();
  }

}
